﻿// PlayStation ELF Converter
// Converts between ELF and PSX-EXE executable formats
// Copyright 2019 Ben Lincoln
// https://www.beneaththewaves.net/
// Based in part on information documented in LameGuy64's elf2x utility
// (http://www.psxdev.net/forum/viewtopic.php?f=75&t=1613)

// This file is part of PlayStation ELF Converter.

// PlayStation ELF Converter is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// PlayStation ELF Converter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with PlayStation ELF Converter (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

using BenLincoln.ExecutableFiles;

namespace PlayStationELFConverter
{
    enum OpMode
    {
        ELFtoPSXEXE,
        PSXEXEtoELF,
        DisplayFileInformation
    }

    enum FileType
    {
        Unknown,
        ELF,
        PSXEXE
    }

    class Program
    {
        protected const string ProgramName = "PlayStation ELF Converter";
        protected static string Copyright = "Copyright 2019 Ben Lincoln" + Environment.NewLine + "Based in part on information documented in LameGuy64's elf2x utility (http://www.psxdev.net/forum/viewtopic.php?f=75&t=1613)";
        protected const string InfoURL = "https://www.beneaththewaves.net/Software/";
        protected const int ReadBufferLineCount = 10000;

        protected static uint ParseUIntParam(string inputParam)
        {
            string input = inputParam.ToLower();
            if (input.Contains("0x"))
            {
                try
                {
                    uint result = UInt32.Parse(input, System.Globalization.NumberStyles.HexNumber);
                    return result;
                }
                catch (Exception ex)
                {
                    ExitWithError(String.Format("Could not parse '{0}' as an unsigned, 32-bit integer: {1}", inputParam, ex.Message), 1);
                }
            }
            else
            {
                try
                {
                    uint result = UInt32.Parse(input, System.Globalization.NumberStyles.Integer);
                    return result;
                }
                catch (Exception ex)
                {
                    ExitWithError(String.Format("Could not parse '{0}' as an unsigned, 32-bit integer: {1}", inputParam, ex.Message), 1);
                }
            }
            throw new Exception("Logical error - contact the developer.");
        }

        static void Main(string[] args)
        {
            if ((args.Length < 2))
            {
                DisplayHelpMessage();
                Environment.Exit(1);
            }

            OpMode mode = OpMode.DisplayFileInformation;
            bool forceOverwrite = false;
            bool quiet = false;
            string inputFilePath = "";
            string outputFilePath = "";

            uint TextSegmentOffset = 0;
            uint TextSegmentAddress = 0;
            uint TextSegmentSize = 0;
            uint DataSegmentOffset = 0;
            uint DataSegmentAddress = 0;
            uint DataSegmentSize = 0;
            uint BSSSegmentOffset = 0;
            uint BSSSegmentAddress = 0;
            uint BSSSegmentSize = 0;
            uint StackSegmentAddress = 0;
            uint StackSegmentSize = 0;

            bool textSegmentExplicitlyDefined = false;
            bool dataSegmentExplicitlyDefined = false;
            bool bssSegmentExplicitlyDefined = false;
            bool stackSegmentExplicitlyDefined = false;


            ArrayList arguments = new ArrayList();
            for (int i = 0; i < args.Length; i++)
            {
                // if the argument is one with a second component, catch that now
                string nextArg = "";
                switch (args[i])
                {
                    case "--text-offset":
                        nextArg = args[i + 1];
                        TextSegmentAddress = ParseUIntParam(nextArg);
                        textSegmentExplicitlyDefined = true;
                        i++;
                        break;
                    case "--text-address":
                        nextArg = args[i + 1];
                        TextSegmentAddress = ParseUIntParam(nextArg);
                        textSegmentExplicitlyDefined = true;
                        i++;
                        break;
                    case "--text-size":
                        nextArg = args[i + 1];
                        TextSegmentSize = ParseUIntParam(nextArg);
                        textSegmentExplicitlyDefined = true;
                        i++;
                        break;
                    case "--data-offset":
                        nextArg = args[i + 1];
                        DataSegmentAddress = ParseUIntParam(nextArg);
                        dataSegmentExplicitlyDefined = true;
                        i++;
                        break;
                    case "--data-address":
                        nextArg = args[i + 1];
                        DataSegmentAddress = ParseUIntParam(nextArg);
                        dataSegmentExplicitlyDefined = true;
                        i++;
                        break;
                    case "--data-size":
                        nextArg = args[i + 1];
                        DataSegmentSize = ParseUIntParam(nextArg);
                        dataSegmentExplicitlyDefined = true;
                        i++;
                        break;
                    case "--bss-offset":
                        nextArg = args[i + 1];
                        BSSSegmentAddress = ParseUIntParam(nextArg);
                        bssSegmentExplicitlyDefined = true;
                        i++;
                        break;
                    case "--bss-address":
                        nextArg = args[i + 1];
                        BSSSegmentAddress = ParseUIntParam(nextArg);
                        bssSegmentExplicitlyDefined = true;
                        i++;
                        break;
                    case "--bss-size":
                        nextArg = args[i + 1];
                        BSSSegmentSize = ParseUIntParam(nextArg);
                        bssSegmentExplicitlyDefined = true;
                        i++;
                        break;
                    case "--stack-address":
                        nextArg = args[i + 1];
                        StackSegmentAddress = ParseUIntParam(nextArg);
                        stackSegmentExplicitlyDefined = true;
                        i++;
                        break;
                    case "--stack-size":
                        nextArg = args[i + 1];
                        StackSegmentSize = ParseUIntParam(nextArg);
                        stackSegmentExplicitlyDefined = true;
                        i++;
                        break;
                    default:
                        arguments.Add(args[i]);
                        break;
                }
            }

            ArrayList positionalArgs = new ArrayList();

            foreach (string arg in arguments)
            {
                if ((arg == "-?") || (arg == "-h") || (arg == "--help"))
                {
                    DisplayHelpMessage();
                    Environment.Exit(0);
                }
                bool handled = false;
                if (arg == "--info")
                {
                    mode = OpMode.DisplayFileInformation;
                    handled = true;
                }
                if (arg == "--elf2exe")
                {
                    mode = OpMode.ELFtoPSXEXE;
                    handled = true;
                }
                if (arg == "--exe2elf")
                {
                    mode = OpMode.PSXEXEtoELF;
                    handled = true;
                }
                if (arg == "--quiet")
                {
                    quiet = true;
                    handled = true;
                }
                if (arg == "--overwrite")
                {
                    forceOverwrite = true;
                    handled = true;
                }
                // right now, all positional arguments are file paths
                if (!handled)
                {
                    positionalArgs.Add(arg);
                }
            }

            if (positionalArgs.Count > 0)
            {
                inputFilePath = (string)positionalArgs[0];
            }

            if (positionalArgs.Count > 1)
            {
                outputFilePath = (string)positionalArgs[1];
            }

            if (inputFilePath == "")
            {
                ExitWithError("No input file path was specified.", 1);
            }

            if (!File.Exists(inputFilePath))
            {
                ExitWithError("The specified input file does not exist.", 1);
            }

            if ((mode == OpMode.ELFtoPSXEXE) || (mode == OpMode.PSXEXEtoELF))
            {
                if (outputFilePath == "")
                {
                    ExitWithError("No output file path was specified.", 1);
                }
                if (!forceOverwrite)
                {
                    if (File.Exists(outputFilePath))
                    {
                        ExitWithError("The specified output file already exists. Use the --overwrite option to replace it.", 1);
                    }
                }
            }

            BenLincoln.ExecutableFiles.BinaryFile inputFile = new BenLincoln.ExecutableFiles.BinaryFile();
            BenLincoln.ExecutableFiles.PSXEXE inputFilePSXEXE = new BenLincoln.ExecutableFiles.PSXEXE();
            BenLincoln.ExecutableFiles.ELF inputFileELF = new BenLincoln.ExecutableFiles.ELF();
            string exceptionMessagePSXEXE = "";
            string exceptionMessageELF = "";
            bool parsedFile = false;

            try
            {
                FileStream inputStream = new FileStream(inputFilePath, FileMode.Open, FileAccess.Read);
                BinaryReader inputReader = new BinaryReader(inputStream);

                try
                {
                    inputFilePSXEXE = new BenLincoln.ExecutableFiles.PSXEXE();
                    inputFilePSXEXE.ReadFile(inputReader);
                    inputFile = inputFilePSXEXE;
                    parsedFile = true;
                }
                catch (Exception fe)
                {
                    parsedFile = false;
                    exceptionMessagePSXEXE = fe.Message;
                    //ExitWithError(String.Format("Could not parse the input file: {0}", fe.Message), 2);
                }

                if (!parsedFile)
                {
                    try
                    {
                        inputFileELF = new BenLincoln.ExecutableFiles.ELF();
                        inputFileELF.ReadFile(inputReader);
                        inputFile = inputFileELF;
                        parsedFile = true;
                    }
                    catch (Exception fe)
                    {
                        parsedFile = false;
                        exceptionMessageELF = fe.Message;
                        //ExitWithError(String.Format("Could not parse the input file: {0}", fe.Message), 2);
                    }
                }
            }
            catch (Exception ex)
            {
                ExitWithError(String.Format("Could not read the input file: {0}", ex.Message), 2);
            }

            if (!parsedFile)
            {
                string failureMessage = "Could not parse the input file as any filetype recognized by this program.";
                if (exceptionMessagePSXEXE != "")
                {
                    failureMessage = String.Format("{0}{1}{2}{3}", failureMessage, Environment.NewLine, "Parsed as PSX-EXE: ", exceptionMessagePSXEXE);
                }
                if (exceptionMessageELF != "")
                {
                    failureMessage = String.Format("{0}{1}{2}{3}", failureMessage, Environment.NewLine, "Parsed as ELF: ", exceptionMessageELF);
                }
                ExitWithError(failureMessage, 2);
            }

            if ((!quiet) || (mode == OpMode.DisplayFileInformation))
            {
                Console.WriteLine("-----------------------------------");
                Console.WriteLine(String.Format("Input file: {0}", inputFilePath));
                Console.WriteLine(inputFile.GetFileInfoString());
                Console.WriteLine("-----------------------------------");
            }

            if ((mode == OpMode.ELFtoPSXEXE) || (mode == OpMode.PSXEXEtoELF))
            {
                string checkOutputText = "";
                if (mode == OpMode.ELFtoPSXEXE)
                {
                    if (inputFile.FileTypeName != "ELF")
                    {
                        ExitWithError("The operating mode was specified as ELF to PSX-EXE, but the input file could not be parsed as an ELF.", 3);
                    }
                }
                if (mode == OpMode.PSXEXEtoELF)
                {
                    if (inputFile.FileTypeName != "PSX-EXE")
                    {
                        ExitWithError("The operating mode was specified as PSX-EXE to ELF, but the input file could not be parsed as a PSX-EXE.", 3);
                    }
                    BenLincoln.ExecutableFiles.ELF outputFile = new BenLincoln.ExecutableFiles.ELF();
                    outputFile.SetFromPSXEXE(inputFilePSXEXE);
                    try
                    {
                        FileStream outStream = new FileStream(outputFilePath, FileMode.Create, FileAccess.ReadWrite);
                        BinaryWriter outWriter = new BinaryWriter(outStream);
                        outputFile.WriteFile(outWriter);
                        outWriter.Close();
                        outStream.Close();
                    }
                    catch (Exception ex)
                    {
                        ExitWithError(String.Format("An error occurred while converting the PSX-EXE file to an ELF: {0}", ex.Message), 4);
                    }
                    // sanity check
                    try
                    {
                        FileStream checkOutputStream = new FileStream(outputFilePath, FileMode.Open, FileAccess.Read);
                        BinaryReader checkOutputReader = new BinaryReader(checkOutputStream);
                        try
                        {
                            BenLincoln.ExecutableFiles.ELF checkFile = new BenLincoln.ExecutableFiles.ELF();
                            checkFile.ReadFile(checkOutputReader);
                            checkOutputText = checkFile.GetFileInfoString();
                        }
                        catch (Exception fe)
                        {
                            ExitWithError(String.Format("Could not parse the output file after creating it: {0}", fe.Message), 5);
                        }
                    }
                    catch (Exception ex)
                    {
                        ExitWithError(String.Format("Could not read the output file after creating it: {0}", ex.Message), 6);
                    }
                }
                Console.WriteLine("-----------------------------------");
                Console.WriteLine(String.Format("Output file: {0}", outputFilePath));
                Console.WriteLine(checkOutputText);
                Console.WriteLine("-----------------------------------");
            }

            Console.WriteLine("");
        }

        static void DisplayHelpMessage()
        {
            DisplayBanner();

            Assembly asm = Assembly.GetExecutingAssembly();
            Stream inStream = asm.GetManifestResourceStream("PlayStationELFConverter.HelpMessage.txt");
            StreamReader inReader = new StreamReader(inStream);
            Console.Write(inReader.ReadToEnd());
            Console.WriteLine();
            inReader.Close();
            inStream.Close();
        }

        static void DisplayBanner()
        {
            Assembly currentAssembly = Assembly.GetExecutingAssembly();

            AssemblyName assemblyName = currentAssembly.GetName();

            int versionMajor = assemblyName.Version.Major;
            int versionMinor = assemblyName.Version.Minor;

            Console.WriteLine(string.Format("{0} {1}.{2}", ProgramName, versionMajor, versionMinor));
            Console.WriteLine(Copyright);
            Console.WriteLine(InfoURL);
            Console.WriteLine(Environment.NewLine);
        }

        protected static void ExitWithError(string message, int exitCode)
        {
            DisplayBanner();
            Console.WriteLine(message);
            System.Environment.ExitCode = exitCode;
            Environment.Exit(exitCode);
        }
    }
}
